/* $RCSfile: curses.mus,v $$Revision: 4.0.1.1 $$Date: 91/11/05 19:06:19 $
 *
 * $Log:	curses.mus,v $
 * Revision 4.0.1.1  91/11/05  19:06:19  lwall
 * patch11: usub/curses.mus now supports SysV curses
 * 
 * Revision 4.0  91/03/20  01:56:13  lwall
 * 4.0 baseline.
 * 
 * Revision 3.0.1.1  90/08/09  04:05:21  lwall
 * patch19: Initial revision
 * 
 */

#include "EXTERN.h"
#include "perl.h"

char *savestr();

#undef bool
#include <curses.h>

#ifndef A_UNDERLINE
#define NOSETATTR
#define A_STANDOUT  0x0200
#define A_UNDERLINE 0x0100
#define A_REVERSE   0x0200
#define A_BLINK     0x0400
#define A_BOLD      0x0800
#define A_ALTCHARSET 0x1000    
#define A_DIM	    0x2000
#define A_NORMAL    0
#endif

#ifdef USG
static char *tcbuf = NULL;
#endif

#ifdef NOSETATTR
static unsigned curattr = A_NORMAL;
#endif

typedef char chtype;

static enum uservars {
    UV_curscr,
    UV_stdscr,
    UV_ttytype,
    UV_LINES,
    UV_COLS,
    UV_ERR,
    UV_OK,
#ifdef BSD
    UV_Def_term,
    UV_My_term,
#endif    
    UV_A_STANDOUT,
    UV_A_UNDERLINE,
    UV_A_REVERSE,
    UV_A_BLINK,
    UV_A_DIM,
    UV_A_BOLD,
    UV_A_NORMAL,
};

static enum usersubs {
    US_addch,
    US_waddch,
    US_addstr,
    US_waddstr,
    US_box,
    US_clear,
    US_wclear,
    US_clearok,
    US_clrtobot,
    US_wclrtobot,
    US_clrtoeol,
    US_wclrtoeol,
    US_delch,
    US_wdelch,
    US_deleteln,
    US_wdeleteln,
    US_erase,
    US_werase,
    US_idlok,
    US_insch,
    US_winsch,
    US_insertln,
    US_winsertln,
    US_move,
    US_wmove,
    US_overlay,
    US_overwrite,
    US_refresh,
    US_wrefresh,
    US_standout,
    US_wstandout,
    US_standend,
    US_wstandend,
    US_cbreak,
    US_nocbreak,
    US_echo,
    US_noecho,
    US_getch,
    US_wgetch,
    US_getstr,
    US_wgetstr,
    US_raw,
    US_noraw,
    US_baudrate,
    US_delwin,
    US_endwin,
    US_erasechar,
    US_getyx,
    US_inch,
    US_winch,
    US_initscr,
    US_killchar,
    US_leaveok,
    US_longname,
    US_mvwin,
    US_newwin,
    US_nl,
    US_nonl,
    US_scrollok,
    US_subwin,
    US_touchline,
    US_touchwin,
    US_unctrl,
    US_gettmode,
    US_mvcur,
    US_scroll,
    US_savetty,
    US_resetty,
    US_setterm,
    US_attroff,
    US_wattroff,
    US_attron,
    US_wattron,
    US_attrset,
    US_wattrset,
#ifdef CURSEFMT
    US_printw, /* remove */
    US_wprintw, /* remove */
    US_scanw, 	/* delete */
    US_wscanw, 	/* delete */
#endif
    US_getcap,
#ifdef BSD
    US_flushok,
    US_fullname,
    US_touchoverlap,
    US_tstp,
    US__putchar,
#endif
    US_mysub,
    US_testcallback,
};

static int usersub();
static int userset();
static int userval();

int
init_curses()
{
    struct ufuncs uf;
    char *filename = "curses.c";

    uf.uf_set = userset;
    uf.uf_val = userval;

#define MAGICVAR(name, ix) uf.uf_index = ix, magicname(name, &uf, sizeof uf)

    MAGICVAR("curscr",	UV_curscr);
    MAGICVAR("stdscr",	UV_stdscr);
    MAGICVAR("ttytype",	UV_ttytype);
    MAGICVAR("LINES",	UV_LINES);
    MAGICVAR("COLS",	UV_COLS);
    MAGICVAR("ERR",	UV_ERR);
    MAGICVAR("OK",	UV_OK);
#ifdef BSD
    MAGICVAR("Def_term",UV_Def_term);
    MAGICVAR("My_term",	UV_My_term);
#endif
    MAGICVAR("A_STANDOUT", UV_A_STANDOUT);
    MAGICVAR("A_UNDERLINE", UV_A_UNDERLINE);
    MAGICVAR("A_REVERSE", UV_A_REVERSE);
    MAGICVAR("A_BLINK", UV_A_BLINK);
    MAGICVAR("A_DIM", UV_A_DIM);
    MAGICVAR("A_BOLD", UV_A_BOLD);
    MAGICVAR("A_NORMAL", UV_A_NORMAL);

    make_usub("addch",		US_addch,	usersub, filename);
    make_usub("waddch",		US_waddch,	usersub, filename);
    make_usub("addstr",		US_addstr,	usersub, filename);
    make_usub("waddstr",	US_waddstr,	usersub, filename);
    make_usub("box",		US_box,		usersub, filename);
    make_usub("clear",		US_clear,	usersub, filename);
    make_usub("wclear",		US_wclear,	usersub, filename);
    make_usub("clearok",	US_clearok,	usersub, filename);
    make_usub("clrtobot",	US_clrtobot,	usersub, filename);
    make_usub("wclrtobot",	US_wclrtobot,	usersub, filename);
    make_usub("clrtoeol",	US_clrtoeol,	usersub, filename);
    make_usub("wclrtoeol",	US_wclrtoeol,	usersub, filename);
    make_usub("delch",		US_delch,	usersub, filename);
    make_usub("wdelch",		US_wdelch,	usersub, filename);
    make_usub("deleteln",	US_deleteln,	usersub, filename);
    make_usub("wdeleteln",	US_wdeleteln,	usersub, filename);
    make_usub("erase",		US_erase,	usersub, filename);
    make_usub("werase",		US_werase,	usersub, filename);
    make_usub("idlok",		US_idlok,	usersub, filename);
    make_usub("insch",		US_insch,	usersub, filename);
    make_usub("winsch",		US_winsch,	usersub, filename);
    make_usub("insertln",	US_insertln,	usersub, filename);
    make_usub("winsertln",	US_winsertln,	usersub, filename);
    make_usub("move",		US_move,	usersub, filename);
    make_usub("wmove",		US_wmove,	usersub, filename);
    make_usub("overlay",	US_overlay,	usersub, filename);
    make_usub("overwrite",	US_overwrite,	usersub, filename);
    make_usub("refresh",	US_refresh,	usersub, filename);
    make_usub("wrefresh",	US_wrefresh,	usersub, filename);
    make_usub("standout",	US_standout,	usersub, filename);
    make_usub("wstandout",	US_wstandout,	usersub, filename);
    make_usub("standend",	US_standend,	usersub, filename);
    make_usub("wstandend",	US_wstandend,	usersub, filename);
    make_usub("cbreak",		US_cbreak,	usersub, filename);
    make_usub("nocbreak",	US_nocbreak,	usersub, filename);
    make_usub("echo",		US_echo,	usersub, filename);
    make_usub("noecho",		US_noecho,	usersub, filename);
    make_usub("getch",		US_getch,	usersub, filename);
    make_usub("wgetch",		US_wgetch,	usersub, filename);
    make_usub("getstr",		US_getstr,	usersub, filename);
    make_usub("wgetstr",	US_wgetstr,	usersub, filename);
    make_usub("raw",		US_raw,		usersub, filename);
    make_usub("noraw",		US_noraw,	usersub, filename);
    make_usub("baudrate",	US_baudrate,	usersub, filename);
    make_usub("delwin",		US_delwin,	usersub, filename);
    make_usub("endwin",		US_endwin,	usersub, filename);
    make_usub("erasechar",	US_erasechar,	usersub, filename);
    make_usub("getyx",		US_getyx,	usersub, filename);
    make_usub("inch",		US_inch,	usersub, filename);
    make_usub("winch",		US_winch,	usersub, filename);
    make_usub("initscr",	US_initscr,	usersub, filename);
    make_usub("killchar",	US_killchar,	usersub, filename);
    make_usub("leaveok",	US_leaveok,	usersub, filename);
    make_usub("longname",	US_longname,	usersub, filename);
    make_usub("mvwin",		US_mvwin,	usersub, filename);
    make_usub("newwin",		US_newwin,	usersub, filename);
    make_usub("nl",		US_nl,		usersub, filename);
    make_usub("nonl",		US_nonl,	usersub, filename);
    make_usub("scrollok",	US_scrollok,	usersub, filename);
    make_usub("subwin",		US_subwin,	usersub, filename);
    make_usub("touchline",	US_touchline,	usersub, filename);
    make_usub("touchwin",	US_touchwin,	usersub, filename);
    make_usub("unctrl",		US_unctrl,	usersub, filename);
    make_usub("gettmode",	US_gettmode,	usersub, filename);
    make_usub("mvcur",		US_mvcur,	usersub, filename);
    make_usub("scroll",		US_scroll,	usersub, filename);
    make_usub("savetty",	US_savetty,	usersub, filename);
    make_usub("resetty",	US_resetty,	usersub, filename);
    make_usub("setterm",	US_setterm,	usersub, filename);
    make_usub("getcap",		US_getcap,	usersub, filename);
    make_usub("attroff",        US_attroff,	usersub, filename);
    make_usub("wattroff",       US_wattroff,	usersub, filename);
    make_usub("attron",         US_attron,	usersub, filename);
    make_usub("wattron",        US_wattron,	usersub, filename);
    make_usub("attrset",        US_attrset,	usersub, filename);
    make_usub("wattrset",       US_wattrset,	usersub, filename);
#ifdef CURSEFMT
    make_usub("printw",		US_printw,	usersub, filename);
    make_usub("wprintw",	US_wprintw,	usersub, filename);
    make_usub("scanw",		US_scanw,	usersub, filename);
    make_usub("wscanw",		US_wscanw,	usersub, filename);
#endif
#ifdef BSD
    make_usub("flushok",	US_flushok,	usersub, filename);
    make_usub("fullname",	US_fullname,	usersub, filename);
    make_usub("touchoverlap",	US_touchoverlap,usersub, filename);
    make_usub("tstp",		US_tstp,	usersub, filename);
    make_usub("_putchar",	US__putchar,	usersub, filename);
#endif
    make_usub("testcallback",	US_testcallback,usersub, filename);
  };
  
#ifdef USG
static char
*getcap(cap)
register char *cap;
{
    static char nocaperr[] = "Cannot read termcap entry.";

    extern char *tgetstr();

    if (tcbuf == NULL) {
	if ((tcbuf = malloc(1024)) == NULL) {
	    fatal(nocaperr);
	}
	if (tgetent(tcbuf, ttytype) == -1) {
	    fatal(nocaperr);
	}
    }

    return (tgetstr(cap, NULL));
}
#endif

#ifdef NOSETATTR
#define attron(attr)    wattron(stdscr, attr)
#define attroff(attr)   wattroff(stdscr, attr)
#define attset(attr)    wattset(stdscr, attr)

int
wattron(win, attr)
WINDOW *win;
chtype attr;
{
    curattr |= attr;
    if (curattr & A_STANDOUT) {
	return(wstandout(win));
    } else {
	return(wstandend(win));
    }
}

int
wattroff(win, attr)
WINDOW *win;
chtype attr;
{
    curattr &= (~attr);
    if (curattr & A_STANDOUT) {
	return(wstandout(win));
    } else {
	return(wstandend(win));
    }
}

int
attrset(attr)
chtype attr;
{
    curattr = attr;
    if (curattr & A_STANDOUT) {
	return(standout(win));
    } else {
	return(standend(win));
    }
}

int
wattrset(win, attr)
WINDOW *win;
chtype attr;
{
    curattr = attr;
    if (curattr & A_STANDOUT) {
	return(wstandout(win));
    } else {
	return(wstandend(win));
    }
}

#endif
    
static int
usersub(ix, sp, items)
int ix;
register int sp;
register int items;
{
    STR **st = stack->ary_array + sp;
    register int i;
    register char *tmps;
    register STR *Str;		/* used in str_get and str_gnum macros */

    switch (ix) {
    case US_addch:
	if (items != 1)
	    fatal("Usage: &addch($ch)");
	else {
	    int retval;
	    char	ch =		(char)		str_gnum(st[1]);

	    retval = addch(ch);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_waddch:
	if (items != 2)
	    fatal("Usage: &waddch($win, $ch)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    char	ch =		(char)		str_gnum(st[2]);

	    retval = waddch(win, ch);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_addstr:
	if (items != 1)
	    fatal("Usage: &addstr($str)");
	else {
	    int retval;
	    char*	str =		(char*)		str_get(st[1]);

	    retval = addstr(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_waddstr:
	if (items != 2)
	    fatal("Usage: &waddstr($win, $str)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    char*	str =		(char*)		str_get(st[2]);

	    retval = waddstr(win, str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_box:
	if (items != 3)
	    fatal("Usage: &box($win, $vert, $hor)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    char	vert =		(char)		str_gnum(st[2]);
	    char	hor =		(char)		str_gnum(st[3]);

	    retval = box(win, vert, hor);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_clear:
	if (items != 0)
	    fatal("Usage: &clear()");
	else {
	    int retval;

	    retval = clear();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wclear:
	if (items != 1)
	    fatal("Usage: &wclear($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wclear(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_clearok:
	if (items != 2)
	    fatal("Usage: &clearok($win, $boolf)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    bool	boolf =		(bool)		str_gnum(st[2]);

	    retval = clearok(win, boolf);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_clrtobot:
	if (items != 0)
	    fatal("Usage: &clrtobot()");
	else {
	    int retval;

	    retval = clrtobot();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wclrtobot:
	if (items != 1)
	    fatal("Usage: &wclrtobot($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wclrtobot(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_clrtoeol:
	if (items != 0)
	    fatal("Usage: &clrtoeol()");
	else {
	    int retval;

	    retval = clrtoeol();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wclrtoeol:
	if (items != 1)
	    fatal("Usage: &wclrtoeol($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wclrtoeol(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_delch:
	if (items != 0)
	    fatal("Usage: &delch()");
	else {
	    int retval;

	    retval = delch();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wdelch:
	if (items != 1)
	    fatal("Usage: &wdelch($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wdelch(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_deleteln:
	if (items != 0)
	    fatal("Usage: &deleteln()");
	else {
	    int retval;

	    retval = deleteln();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wdeleteln:
	if (items != 1)
	    fatal("Usage: &wdeleteln($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wdeleteln(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_erase:
	if (items != 0)
	    fatal("Usage: &erase()");
	else {
	    int retval;

	    retval = erase();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_werase:
	if (items != 1)
	    fatal("Usage: &werase($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = werase(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_idlok:
	if (items != 2)
	    fatal("Usage: &idlok($win, $boolf)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    bool	boolf =		(bool)		str_gnum(st[2]);

	    retval = idlok(win, boolf);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_insch:
	if (items != 1)
	    fatal("Usage: &insch($c)");
	else {
	    int retval;
	    char	c =		(char)		str_gnum(st[1]);

	    retval = insch(c);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_winsch:
	if (items != 2)
	    fatal("Usage: &winsch($win, $c)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    char	c =		(char)		str_gnum(st[2]);

	    retval = winsch(win, c);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_insertln:
	if (items != 0)
	    fatal("Usage: &insertln()");
	else {
	    int retval;

	    retval = insertln();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_winsertln:
	if (items != 1)
	    fatal("Usage: &winsertln($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = winsertln(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_move:
	if (items != 2)
	    fatal("Usage: &move($y, $x)");
	else {
	    int retval;
	    int		y =		(int)		str_gnum(st[1]);
	    int		x =		(int)		str_gnum(st[2]);

	    retval = move(y, x);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wmove:
	if (items != 3)
	    fatal("Usage: &wmove($win, $y, $x)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    int		y =		(int)		str_gnum(st[2]);
	    int		x =		(int)		str_gnum(st[3]);

	    retval = wmove(win, y, x);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_overlay:
	if (items != 2)
	    fatal("Usage: &overlay($win1, $win2)");
	else {
	    int retval;
	    WINDOW*	win1 =		*(WINDOW**)	str_get(st[1]);
	    WINDOW*	win2 =		*(WINDOW**)	str_get(st[2]);

	    retval = overlay(win1, win2);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_overwrite:
	if (items != 2)
	    fatal("Usage: &overwrite($win1, $win2)");
	else {
	    int retval;
	    WINDOW*	win1 =		*(WINDOW**)	str_get(st[1]);
	    WINDOW*	win2 =		*(WINDOW**)	str_get(st[2]);

	    retval = overwrite(win1, win2);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_refresh:
	if (items != 0)
	    fatal("Usage: &refresh()");
	else {
	    int retval;

	    retval = refresh();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wrefresh:
	if (items != 1)
	    fatal("Usage: &wrefresh($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wrefresh(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_standout:
	if (items != 0)
	    fatal("Usage: &standout()");
	else {
	    int retval;

	    retval = standout();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wstandout:
	if (items != 1)
	    fatal("Usage: &wstandout($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wstandout(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_standend:
	if (items != 0)
	    fatal("Usage: &standend()");
	else {
	    int retval;

	    retval = standend();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wstandend:
	if (items != 1)
	    fatal("Usage: &wstandend($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = wstandend(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_cbreak:
	if (items != 0)
	    fatal("Usage: &cbreak()");
	else {
	    int retval;

	    retval = cbreak();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_nocbreak:
	if (items != 0)
	    fatal("Usage: &nocbreak()");
	else {
	    int retval;

	    retval = nocbreak();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_echo:
	if (items != 0)
	    fatal("Usage: &echo()");
	else {
	    int retval;

	    retval = echo();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_noecho:
	if (items != 0)
	    fatal("Usage: &noecho()");
	else {
	    int retval;

	    retval = noecho();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_getch:
	if (items != 0)
	    fatal("Usage: &getch()");
	else {
	    int retval;
	    char retch;

	    retval = getch();
	    if (retval == EOF)
		st[0] = &str_undef;
	    else {
		retch = retval;
		str_nset(st[0], &retch, 1);
	    }
	}
	return sp;

    case US_wgetch:
	if (items != 1)
	    fatal("Usage: &wgetch($win)");
	else {
	    int retval;
	    char retch;
	    WINDOW*     win =           *(WINDOW**)     str_get(st[1]);

	    retval = wgetch(win);
	    if (retval == EOF)
		st[0] = &str_undef;
	    else {
		retch = retval;
		str_nset(st[0], &retch, 1);
	    }
	}
	return sp;

    case US_getstr:
	if (items != 1)
	    fatal("Usage: &getstr($str)");
	else {
	    int retval;
	    char	str[133];

	    retval = getstr(str);
	    str_numset(st[0], (double) retval);
	    str_set(st[1], (char*) str);
	}
	return sp;

    case US_wgetstr:
	if (items != 2)
	    fatal("Usage: &wgetstr($win, $str)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    char	str[133];

	    retval = wgetstr(win, str);
	    str_numset(st[0], (double) retval);
	    str_set(st[2], (char*) str);
	}
	return sp;

    case US_raw:
	if (items != 0)
	    fatal("Usage: &raw()");
	else {
	    int retval;

	    retval = raw();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_noraw:
	if (items != 0)
	    fatal("Usage: &noraw()");
	else {
	    int retval;

	    retval = noraw();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_baudrate:
	if (items != 0)
	    fatal("Usage: &baudrate()");
	else {
	    int retval;

	    retval = baudrate();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_delwin:
	if (items != 1)
	    fatal("Usage: &delwin($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = delwin(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_endwin:
	if (items != 0)
	    fatal("Usage: &endwin()");
	else {
	    int retval;

	    retval = endwin();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_erasechar:
	if (items != 0)
	    fatal("Usage: &erasechar()");
	else {
	    int retval;

	    retval = erasechar();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_getyx:
	if (items != 3)
	    fatal("Usage: &getyx($win, $y, $x)");
	else {
	    int retval;
	    STR*	str =		str_new(0);
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    int		y;
	    int		x;

	    do_sprintf(str, items - 1, st + 1);
	    retval = getyx(win, y, x);
	    str_numset(st[2], (double)y);
	    str_numset(st[3], (double)x);
	    str_numset(st[0], (double) retval);
	    str_free(str);
	}
	return sp;
	
    case US_inch:
	if (items != 0)
	    fatal("Usage: &inch()");
	else {
	    int retval;

	    retval = inch();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_winch:
	if (items != 1)
	    fatal("Usage: &winch($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = winch(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_initscr:
	if (items != 0)
	    fatal("Usage: &initscr()");
	else {
	    WINDOW* retval;

	    retval = initscr();
	    str_nset(st[0], (char*) &retval, sizeof retval);
	}
	return sp;

    case US_killchar:
	if (items != 0)
	    fatal("Usage: &killchar()");
	else {
	    int retval;

	    retval = killchar();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_leaveok:
	if (items != 2)
	    fatal("Usage: &leaveok($win, $boolf)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    bool	boolf =		(bool)		str_gnum(st[2]);

	    retval = leaveok(win, boolf);
	    str_numset(st[0], (double) retval);
	}
	return sp;

#ifdef BSD
    case US_longname:
	if (items != 2)
	    fatal("Usage: &longname($termbuf, $name)");
	else {
	    char* retval;
	    char*	termbuf =	(char*)		str_get(st[1]);
	    char*	name =		(char*)		str_get(st[2]);

	    retval = longname(termbuf, name);
	    str_set(st[0], (char*) retval);
	    str_set(st[2], (char*) name);
	}
	return sp;
#else
    case US_longname:
	if (items != 2)
	    fatal("Usage: &longname($termbug, $name)");
	else {
	    char* retval;
	    char*	termbug =	(char*)		str_get(st[1]);
	    char*	name =		(char*)		str_get(st[2]);

	    retval = longname(termbug, name);
	    str_set(st[0], (char*) retval);
	}
	return sp;
#endif

    case US_mvwin:
	if (items != 3)
	    fatal("Usage: &mvwin($win, $y, $x)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    int		y =		(int)		str_gnum(st[2]);
	    int		x =		(int)		str_gnum(st[3]);

	    retval = mvwin(win, y, x);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_newwin:
	if (items != 4)
	    fatal("Usage: &newwin($lines, $cols, $begin_y, $begin_x)");
	else {
	    WINDOW* retval;
	    int		lines =		(int)		str_gnum(st[1]);
	    int		cols =		(int)		str_gnum(st[2]);
	    int		begin_y =	(int)		str_gnum(st[3]);
	    int		begin_x =	(int)		str_gnum(st[4]);

	    retval = newwin(lines, cols, begin_y, begin_x);
	    str_nset(st[0], (char*) &retval, sizeof retval);
	}
	return sp;

    case US_nl:
	if (items != 0)
	    fatal("Usage: &nl()");
	else {
	    int retval;

	    retval = nl();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_nonl:
	if (items != 0)
	    fatal("Usage: &nonl()");
	else {
	    int retval;

	    retval = nonl();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_scrollok:
	if (items != 2)
	    fatal("Usage: &scrollok($win, $boolf)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    bool	boolf =		(bool)		str_gnum(st[2]);

	    retval = scrollok(win, boolf);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_subwin:
	if (items != 5)
	    fatal("Usage: &subwin($win, $lines, $cols, $begin_y, $begin_x)");
	else {
	    WINDOW* retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    int		lines =		(int)		str_gnum(st[2]);
	    int		cols =		(int)		str_gnum(st[3]);
	    int		begin_y =	(int)		str_gnum(st[4]);
	    int		begin_x =	(int)		str_gnum(st[5]);

	    retval = subwin(win, lines, cols, begin_y, begin_x);
	    str_nset(st[0], (char*) &retval, sizeof retval);
	}
	return sp;

    case US_touchline:
	if (items != 4)
	    fatal("Usage: &touchline($win, $y, $startx, $endx)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    int		y =		(int)		str_gnum(st[2]);
	    int		startx =	(int)		str_gnum(st[3]);
	    int		endx =		(int)		str_gnum(st[4]);

	    retval = touchline(win, y, startx, endx);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_touchwin:
	if (items != 1)
	    fatal("Usage: &touchwin($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = touchwin(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_unctrl:
	if (items != 1)
	    fatal("Usage: &unctrl($ch)");
	else {
	    char* retval;
	    char	ch =		(char)		str_gnum(st[1]);

	    retval = unctrl(ch);
	    str_set(st[0], (char*) retval);
	}
	return sp;

    case US_gettmode:
	if (items != 0)
	    fatal("Usage: &gettmode()");
	else {
	    int retval;

	    retval = gettmode();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_mvcur:
	if (items != 4)
	    fatal("Usage: &mvcur($lasty, $lastx, $newy, $newx)");
	else {
	    int retval;
	    int		lasty =		(int)		str_gnum(st[1]);
	    int		lastx =		(int)		str_gnum(st[2]);
	    int		newy =		(int)		str_gnum(st[3]);
	    int		newx =		(int)		str_gnum(st[4]);

	    retval = mvcur(lasty, lastx, newy, newx);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_scroll:
	if (items != 1)
	    fatal("Usage: &scroll($win)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    retval = scroll(win);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_savetty:
	if (items != 0)
	    fatal("Usage: &savetty()");
	else {
	    int retval;

	    retval = savetty();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_resetty:
	if (items != 0)
	    fatal("Usage: &resetty()");
	else {
	    int retval = 1;

	    (void)resetty();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_setterm:
	if (items != 1)
	    fatal("Usage: &setterm($name)");
	else {
	    int retval;
	    char*	name =		(char*)		str_get(st[1]);

	    retval = setterm(name);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_attroff:
	if (items != 1)
	    fatal("Usage: &attroff($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = attroff(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wattroff:
	if (items != 1)
	    fatal("Usage: &wattroff($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = wattroff(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wattron:
	if (items != 1)
	    fatal("Usage: &wattron($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = wattron(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_attron:
	if (items != 1)
	    fatal("Usage: &attron($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = attron(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_attrset:
	if (items != 1)
	    fatal("Usage: &attrset($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = attrset(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_wattrset:
	if (items != 1)
	    fatal("Usage: &wattrset($str)");
	else {
	    int retval;
	    chtype	str =		(chtype)	str_gnum(st[1]);

	    retval = wattrset(str);
	    str_numset(st[0], (double) retval);
	}
	return sp;

#ifdef	CURSEFMT
    case US_printw:
	if (items < 1)
	    fatal("Usage: &printw($fmt, $arg1, $arg2, ... )");
	else {
	    int retval;
	    STR*	str =		str_new(0);

	    do_sprintf(str, items - 1, st + 1);
	    retval = addstr(str->str_ptr);
	    str_numset(st[0], (double) retval);
	    str_free(str);
	}
	return sp;

    case US_wprintw:
	if (items < 2)
	    fatal("Usage: &wprintw($win, $fmt, $arg1, $arg2, ... )");
	else {
	    int retval;
	    STR*	str =		str_new(0);
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);

	    do_sprintf(str, items - 1, st + 1);
	    retval = waddstr(win, str->str_ptr);
	    str_numset(st[0], (double) retval);
	    str_free(str);
	}
	return sp;

#endif

    case US_getcap:
	if (items != 1)
	    fatal("Usage: &getcap($str)");
	else {
	    char* retval;
	    char*	str =		(char*)		str_get(st[1]);

	    retval = getcap(str);
	    str_set(st[0], (char*) retval);
	}
	return sp;

#ifdef BSD
    case US_flushok:
	if (items != 2)
	    fatal("Usage: &flushok($win, $boolf)");
	else {
	    int retval;
	    WINDOW*	win =		*(WINDOW**)	str_get(st[1]);
	    bool	boolf =		(bool)		str_gnum(st[2]);

	    retval = flushok(win, boolf);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_fullname:
	if (items != 2)
	    fatal("Usage: &fullname($termbuf, $name)");
	else {
	    int retval;
	    char*	termbuf =	(char*)		str_get(st[1]);
	    char*	name =		(char*)		str_get(st[2]);

	    retval = fullname(termbuf, name);
	    str_numset(st[0], (double) retval);
	    str_set(st[2], (char*) name);
	}
	return sp;

    case US_touchoverlap:
	if (items != 2)
	    fatal("Usage: &touchoverlap($win1, $win2)");
	else {
	    int retval;
	    WINDOW*	win1 =		*(WINDOW**)	str_get(st[1]);
	    WINDOW*	win2 =		*(WINDOW**)	str_get(st[2]);

	    retval = touchoverlap(win1, win2);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tstp:
	if (items != 0)
	    fatal("Usage: &tstp()");
	else {
	    int retval;

	    retval = tstp();
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US__putchar:
	if (items != 1)
	    fatal("Usage: &_putchar($ch)");
	else {
	    int retval;
	    char	ch =		(char)		str_gnum(st[1]);

	    retval = _putchar(ch);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_testcallback:
	sp = callback("callback", sp + items, curcsv->wantarray, 1, items);
	break;

#endif

    default:
	fatal("Unimplemented user-defined subroutine");
    }
    return sp;
}

static int
userval(ix, str)
int ix;
STR *str;
{
    switch (ix) {
    case UV_COLS:
	str_numset(str, (double)COLS);
	break;
    case UV_ERR:
	str_numset(str, (double)ERR);
	break;
    case UV_LINES:
	str_numset(str, (double)LINES);
	break;
    case UV_OK:
	str_numset(str, (double)OK);
	break;
    case UV_curscr:
	str_nset(str, &curscr, sizeof(WINDOW*));
	break;
    case UV_stdscr:
	str_nset(str, &stdscr, sizeof(WINDOW*));
	break;
    case UV_ttytype:
	str_set(str, ttytype);
	break;
#ifdef BSD
    case UV_Def_term:
	str_set(str, Def_term);
	break;
    case UV_My_term:
	str_numset(str, (double)My_term);
	break;
#endif
    case UV_A_STANDOUT:
	str_numset(str, (double)A_STANDOUT);
	break;
    case UV_A_UNDERLINE:
	str_numset(str, (double)A_UNDERLINE);
	break;
    case UV_A_REVERSE:
	str_numset(str, (double)A_REVERSE);
	break;
    case UV_A_BLINK:
	str_numset(str, (double)A_BLINK);
	break;
    case UV_A_DIM:
	str_numset(str, (double)A_DIM);
	break;
    case UV_A_BOLD:
	str_numset(str, (double)A_BOLD);
	break;
    case UV_A_NORMAL:
	str_numset(str, (double)A_NORMAL);
	break;
    }
    return 0;
}

static int
userset(ix, str)
int ix;
STR *str;
{
    switch (ix) {
    case UV_COLS:
	COLS = (int)str_gnum(str);
	break;
    case UV_LINES:
	LINES = (int)str_gnum(str);
	break;
    case UV_ttytype:
	strcpy(ttytype, str_get(str));		/* hope it fits */
#ifdef USG
	if (tcbuf != NULL) {
	    free(tcbuf);
	    tcbuf = NULL;
	}
#endif	
	break;
#ifdef BSD
    case UV_Def_term:
	Def_term = savestr(str_get(str));	/* never freed */
	break;
    case UV_My_term:
	My_term = (bool)str_gnum(str);
	break;
#endif
    }
    return 0;
}
